/*
TIFF I/O Library

 Masaomi Nishimura (masaominishimura@riken.jp)

Copyright:(c) Masaomi Nishimura, 2014.


 All right is reserved by Masaomi Nishimura.
This C++ code files are allowed for only primary user of research and educational purposes. Don't use secondary: copy, distribution, diversion, business purpose, and etc.. 

 In no event shall the author be liable to any party for direct, indirect, 
special, incidental, or consequential damage arising out of the use of this 
program and source files. 

*/

#pragma once

#include <string>
#include <fstream>
#include <deque>
#include <map>
#include <climits>
#include "SimpleImage.h"
#include "TiffTags.h"

#ifndef int8
#define int8 signed char
#endif
#ifndef int16
#define int16 signed short int
#endif
#ifndef int32
#define int32 signed int
#endif
#ifndef int64
#define int64 signed long long int
#endif
#ifndef uint8
#define uint8 unsigned char
#endif
#ifndef uint16
#define uint16 unsigned short int
#endif
#ifndef uint32
#define uint32 unsigned int
#endif
#ifndef uint64
#define uint64 unsigned long long int
#endif

#ifdef DEBUG
#include <iostream>
/// write msg & src-file & line-number to STDOUT
/// to enable this function, compile with "DEBUG" definition
#define debug_msg(msg) (std::cout << msg << " @ " << __FILE__ << ":" << __LINE__ << std::endl)
#else
#define debug_msg(msg)
#endif

//namespace tifio {
///////////////////////////////////////////////////////////////////////////////
// definitions of static values
///////////////////////////////////////////////////////////////////////////////
static const char BRICT_STRING[]="Bio-Research Infrastructure Construction Team, VCAD System Research Program, RIKEN, Japan.\0";
static const uint16 SAMPLE_R=0;
static const uint16 SAMPLE_G=1;
static const uint16 SAMPLE_B=2;

///////////////////////////////////////////////////////////////////////////////
/// definition of function reverse bytes array.
/** @param buf pointer to memory buffer.
  * @param size size of one valiable.
  * @param num num of values in the buffer.
  */ //////////////////////////////////////////////////////////////////////////
inline void swapBytes(void* buf,size_t size,size_t num=1){
	if(buf==NULL) return;
	size_t buflen=size*num;
	unsigned char* s=static_cast<unsigned char*>(buf);
	unsigned char c;
	size_t p=0;
	while(p<buflen){
		for(size_t q=0;q<(size>>1);q++){
			c=*(s+p+q);
			*(s+p+q)=*(s+p+size-q-1);
			*(s+p+size-q-1)=c;
		}
		p+=size;
	}
}

///////////////////////////////////////////////////////////////////////////////
/// definitions of class TifIO.
/** import pixels from tiff file to Image / Image3D class.
  * export pixels from Image / Image3D class to tiff file.
  * Image / Image3D class is in the "SimpleImage.h" file.
  */ //////////////////////////////////////////////////////////////////////////
class TifIO
{
	typedef unsigned short tag_t; /// 2-byte unsigned integer
	typedef unsigned int data_t;  /// 4-byte unsigned integer

	protected:
		bool m_fBigEndian; /// is the file written in Big Endian?
		std::deque<std::fstream::pos_type> m_Offsets; /// offsets of pages
		tag_t m_bps; /// bits per sample of the tif file

	public:
		std::string filename; /// file path to read or write data
		std::map<tag_t,CTiffTag> m_ttag; /// tif tag list
		static const data_t IFD_SIZE=12; /// size of a tag

	protected:
		/// read TIFF Sign and check endian
		bool readHeader(std::ifstream& ifp, data_t& head){
			ifp.read(reinterpret_cast<char*>(&head),sizeof(data_t));
			if(head==0x002a4949)
				m_fBigEndian=false;
			else if(head==0x2a004d4d)
				m_fBigEndian=true;
			else
				return false;
			return true;
		}
		/// read 4-byte data from ifp
		void readLong(std::ifstream& ifp, data_t& off){
			ifp.read(reinterpret_cast<char*>(&off),sizeof(data_t));
			if(m_fBigEndian) swapBytes(&off,sizeof(data_t));
		}
		/// read 2-byte data from ifp
		void readShort(std::ifstream& ifp, tag_t& tags){
			ifp.read(reinterpret_cast<char*>(&tags),sizeof(tag_t));
			if(m_fBigEndian) swapBytes(&tags,sizeof(tag_t));
		}
		/// write 4-byte data to ofp
		void writeLong(std::ofstream& ofp,data_t& val){
			ofp.write(reinterpret_cast<char*>(&val),sizeof(data_t));
		}
		/// write 2-byte data to ofp
		void writeShort(std::ofstream& ofp,tag_t& val){
			ofp.write(reinterpret_cast<char*>(&val),sizeof(tag_t));
		}
		/// write CTiffTag 12-byte data to ofp
		void writeTag(std::ofstream& ofp,CTiffTag& tag){
			writeShort(ofp,tag.id);
			writeShort(ofp,tag.type);
			writeLong(ofp,tag.count);
			writeLong(ofp,tag.addr);
		}
		/// read 12-byte data to CTiffTag from ifp
		void readIFD(std::ifstream& ifp);
		/// initialize(clear) this class
		void init(){
			filename="";
			m_ttag.clear();
		}
		/// read a scanline from ifp to buf
		bool readLine(std::ifstream& ifp,double* buf);
		/// write a scanline from buf to ofp
		bool writeLine(std::fstream& ofp,double* buf,data_t len,data_t size,tag_t fmt);
	public:
		TifIO();
		TifIO(const char* in);
		TifIO(const TifIO& rhs);
		~TifIO();

		/// load/saveɎgpt@C(/t)pXݒ肷.
		void setFilename(const char* const in);
		void setFilename(const std::string& in);

		/// filenameŎw肵tift@CpageԂ.
		/** 1ȏ̎R. 1Ȃ2D摜.
		  * TIFFłȂAĂ铙G[̏ꍇ0Ԃ.
		  */
		int getPageNumber();

		/// filenameŎw肵tift@CJ[摜ȂtrueAO[XP[ȂfalseԂ.
		bool isColor(int zID=0);

		/// pathɎw肵tift@CTifIONXœǂݏo\ȂtrueԂ.
		/** pathȗꍇfilenameo̒lgp.
		  */
		bool isApplicable(std::string path="");

		// --------------------------------------------------------------------
		// 2D Image IO
		// --------------------------------------------------------------------
		/// filenameŎw肵tift@CSimpleImage.hImageNXɓǂݍ.
		/** true,s(t@CȂAJ[̏ꍇ)falseԂ.
		  * sb`(px,py),TCY(sx,sy),Pximg.
		  * imgzmemorẙmۂ͊֐ōs.
		  */
		//bool getImage(Image* img);

		/// filenameŎw肵tift@C̃J[摜SimpleImage.hImageNXR,G,BƂɓǂݍ.
		/** true,s(t@CȂAJ[摜łȂ)falseԂ.
		  * sb`(px,py),TCY(sx,sy),Pximg.
		  * imgzmemorẙmۂ͊֐ōs.
		  */
		//bool getColorImage(Image *R,Image *G,Image *B);

		/// filenameŎw肵}`tift@CzIDy[W̉摜SimpleImage.hImageNXɓǂݍ.
		/** true,s(t@CȂAJ[̏ꍇ)falseԂ.
		  * sb`(px,py),TCY(sx,sy),Pximg.
		  * imgzmemorẙmۂ͊֐ōs.
		  */
		bool getImage(Image* image,int zID=0);

		/// filenameŎw肵}`tift@CzIDy[W̉摜SimpleImage.hImageNXR,G,BƂɓǂݍ.
		/** true,s(t@CȂAJ[摜łȂ)falseԂ.
		  * sb`(px,py),TCY(sx,sy),Pximg.
		  * imgzmemorẙmۂ͊֐ōs.
		  */
		bool getColorImage(Image *R,Image *G,Image *B,int zID=0);

		/// SimpleImage.hImageNXin̒gfilenameŎw肵tift@CƂĕۑ.
		/** true,s(t@Copens)falseԂ.
		  * fSaveAsbit=8̏ꍇAsmax,sminɂċPxl0-255Ɋ蓖BYTE(unsigned char)zƂĕۑB
		  * fSaveAsbit=16̏ꍇAsmax,sminɂċPxl0-2^16-1Ɋ蓖unsigned shortzƂĕۑB
		  * fSaveAsbit=64̏ꍇ/w薳̏ꍇdoubleẑ܂ܕۑB
		  */
		bool saveImage(Image *in,double smax=USHRT_MAX,double smin=0,int fSaveAsbit=16);

		/// SimpleImage.hImageNXR,G,B̒gfilenameŎw肵J[tift@CƂĕۑ.
		/** true,s(t@Copens)falseԂ.
		   * fSaveAsbit=8̏ꍇAsmax,sminɂċPxl0-255Ɋ蓖BYTE(unsigned char)zƂĕۑB
		  * fSaveAsbit=16̏ꍇ/w薳̏ꍇsmax,sminɂċPxl0-2^16-1Ɋ蓖unsigned shortzƂĕۑB
		  * fSaveAsbit=64̏ꍇdoubleẑ܂ܕۑB
		  */
		bool saveColorImage(Image *R,Image *G,Image *B,double smax=USHRT_MAX,double smin=0,int fSaveAsbit=16);

		/// SimpleImage.hImageNXin̒gfilenameŎw肵}`tift@CzIDy[WڂƂēւĕۑ.
		/** true,s(t@CopensATCYႢ)falseԂ.
		  */
		bool saveImage(Image *in,int zID);

		/// SimpleImage.hImageNXR,G,B̒gfilenameŎw肵J[}`tift@CzIDy[WڂƂēւĕۑ.
		/** true,s(t@CopensATCYႢ)falseԂ.
		  */
		bool saveColorImage(Image *R,Image *G,Image *B,int zID);

		// --------------------------------------------------------------------
		// 3D Image IO
		// --------------------------------------------------------------------
		/// filenameŎw肵}`tift@CSimpleImage.hImage3DNXɓǂݍ.
		/** s(t@CȂAJ[̏ꍇ)falseԂ.
		  * sb`(px,py,pz),TCY(sx,sy,sz),Pximg.
		  * imgzmemorẙmۂ͊֐ōs.
		  */
		bool getImage(Image3D* image);

		/// filenameŎw肵}`tift@C̃J[摜SimpleImage.hImage3DNXR,G,BƂɓǂݍ.
		/** true,s(t@CȂAJ[摜łȂ)falseԂ.
		  * sb`(px,py,pz),TCY(sx,sy,sz),Pximg.
		  * imgzmemorẙmۂ͊֐ōs.
		  */
		bool getColorImage(Image3D *R,Image3D *G,Image3D *B);

		/// SimpleImage.hImage3DNXin̒gfilenameŎw肵}`tift@CƂĕۑ.
		/** true,s(t@Copens)NULLԂ.
		  * fSaveAsbit=8̏ꍇAsmax,sminɂċPxl0-255Ɋ蓖BYTE(unsigned char)zƂĕۑB
		  * fSaveAsbit=16̏ꍇ/w薳̏ꍇAsmax,sminɂċPxl0-2^16-1Ɋ蓖unsigned shortzƂĕۑB
		  * fSaveAsbit=64̏ꍇdoubleẑ܂ܕۑB
		  */
		bool saveImage(Image3D *in,double smax=USHRT_MAX,double smin=0,int fSaveAsbit=16);

		/// SimpleImage.hImage3DNXR,G,B̒gfilenameŎw肵J[}`tift@CƂĕۑ.
		/** true,s(t@Copens)NULLԂ.
		  * fSaveAsbit=8̏ꍇAsmax,sminɂċPxl0-255Ɋ蓖BYTE(unsigned char)zƂĕۑB
		  * fSaveAsbit=16̏ꍇ/w薳̏ꍇAsmax,sminɂċPxl0-2^16-1Ɋ蓖unsigned shortzƂĕۑB
		  * fSaveAsbit=64̏ꍇdoubleẑ܂ܕۑB
		  */
		bool saveColorImage(Image3D *R,Image3D *G,Image3D *B,double smax=USHRT_MAX,double smin=0,int fSaveAsbit=16);
};

//}
